function [W,b] = trainNetwork(X, Y, W, b, option)
    import TrainingUtils.*
    
    InitialLearnRate    = option.InitialLearnRate;
    LearnRateDropPeriod = option.LearnRateDropPeriod;
    LearnRateDropFactor = option.LearnRateDropFactor;
    BatchSize           = option.BatchSize;
    MaxEpoch            = option.MaxEpoch;
    Shuffle             = option.Shuffle;
    Verbose             = option.Verbose;
    VerboseFrequency    = option.VerboseFrequency;
    ActivationFunction  = option.ActivationFunction;
    LossCriterion       = option.LossCriterion;
    Optimizer           = option.Optimizer;
    
    if Shuffle
        idx = randperm(size(X,2));
        X = X(:,idx);
        Y = Y(:,idx);
    end
    
    if Verbose
       numSample = min(BatchSize, size(X,2));
       preview_X = X(:, 1:numSample); % used for Verbose
       preview_Y = Y(:, 1:numSample);
       
       hfig = figure;

       subplot(1,2,1);
       hfig_loss = animatedline();
       xlabel('iter');ylabel('error');
       title('Gradient Descent Progress')
    end
    
    % cache/velocity used by optimizers such as sgdm, adam
    cache_W = initializeOptimizerCache(Optimizer.solver,W);
    cache_b = initializeOptimizerCache(Optimizer.solver,b);
    
    LearnRates = LearnRateScheduler(InitialLearnRate, MaxEpoch, LearnRateDropPeriod, LearnRateDropFactor);
    NumSample = size(X,2);
    iter = 1;
    for epoch = 1:MaxEpoch
        % get training data of current iteration
        cur_lr = LearnRates(epoch);
        
        indices = generateBatchIndices(NumSample, BatchSize);
        numBatches = length(indices);
        for i = 1: numBatches % mini-batch iteration
            cur_X = X(:, indices{i});
            cur_Y = Y(:, indices{i});
       
            % training
            [loss, cur_Y_hat] = forwardLoss(cur_X,cur_Y,W,b, ActivationFunction, LossCriterion);
            [dLdW, dLdb] = backwardLoss(cur_X,cur_Y,cur_Y_hat,W,b, ActivationFunction, LossCriterion);
            [W, cache_W] = update(Optimizer, W, dLdW, cur_lr, cache_W);
            [b, cache_b] = update(Optimizer, b, dLdb, cur_lr, cache_b);

            % show training progress
            if Verbose && mod(iter, VerboseFrequency) == 0
                preview_Y_hat = predict(preview_X,W,b, ActivationFunction);
                addpoints(hfig_loss, epoch, loss);

                figure(hfig);subplot(1,2,2);
                plot(1:size(preview_Y,2),preview_Y,'ro',...
                       1:size(preview_Y,2),preview_Y_hat,'b+');
                xlabel('x');ylabel('y');title('prediction');

                drawnow;

                fprintf(sprintf("Epoch: %d Iter: %d Loss: %f Learning Rate: %f\n", ...
                    epoch, iter, loss, cur_lr));
            end
            
            iter = iter + 1;
        end
    end
   
    if Verbose
        close(hfig);
    end
end

function lr = LearnRateScheduler(InitialLearnRate, MaxEpoch, LearnRateDropPeriod, LearnRateDropFactor)
% LearnRateScheduler generates a sequence of learning rate using step decay method
    lr = repmat(InitialLearnRate,MaxEpoch,1);
    n_drop = ceil(MaxEpoch/LearnRateDropPeriod);
    for i = 2:n_drop
        pos_l = (i-1)*LearnRateDropPeriod + 1;
        pos_r = min(i*LearnRateDropPeriod, MaxEpoch);
        lr(pos_l:pos_r) = lr(pos_l-1) * LearnRateDropFactor;
    end
end

function indices = generateBatchIndices(n, BatchSize)
% generateBatchIndices generate a cell of indices for selecting mini-batch
% of training data
%
% Output:
%   indices : cell of array
%       X(indices{i}) is the i-th minibatch of training data X
    I = randperm(n);
    numBatch = ceil(n/BatchSize);
    
    indices = cell(numBatch,1);
    for i = 1:numBatch
        l = (i-1) * BatchSize + 1;
        r = min(i * BatchSize, n);
        indices{i} = I(l:r);
    end
end