function [W, cache] = update(Optimizer, W, dLdW, lr, cache)
% update updates network weights W
    solver                = Optimizer.solver;
    GradientClipThreshold = Optimizer.GradientClipThreshold;
    sgdm_momentum         = Optimizer.sgdm_momentum;
    adam_beta1            = Optimizer.adam_beta1;
    adam_beta2            = Optimizer.adam_beta2;
    rms_decayrate         = Optimizer.rms_decayrate;
    
    OptimizerList = struct(...
        'sgd', @(dW) update_sgd(W, dW, lr, cache),...
        'sgdm', @(dW) update_sgdm(W, dW, lr, sgdm_momentum, cache),...
        'adam', @(dW) update_adam(W, dW, lr, adam_beta1, adam_beta2, cache),...
        'rmsprop', @(dW) update_rmsprop(W, dW, lr, rms_decayrate, cache),...
        'adagrad', @(dW) update_adagrad(W, dW, lr, cache));
    update_ = OptimizerList.(lower(solver));
    
    [W, cache] = update_(GradientClip(dLdW, GradientClipThreshold));
end

function dLdW = GradientClip(dLdW, threshold)
    dLdW = min(max(dLdW, -threshold),threshold);
end

function [W, cache] = update_sgd(W, dLdW, lr, cache)
% update_sgd updates network weights W using gradient descent
    W = W - lr * dLdW;
end

function [W, cache] = update_sgdm(W, dLdW, lr, mom, cache)
% update_sgdm updates network weights W using gradient descent with nesterov momentum method
    cache_prev = cache;
    cache = mom * cache - lr * dLdW;
    W = W - mom * cache_prev + (1 + mom) * cache;
end

function [W, cache] = update_adam(W, dLdW, lr, beta1, beta2, cache)
% update_adam updates network weights W using ADAM method
    cache{1} = beta1 * cache{1} + (1-beta1) * dLdW;
    cache{2} = beta2 * cache{2} + (1-beta2) * dLdW.^2;
    W = W - lr * cache{1} ./ (sqrt(cache{2}) + 1e-10);
end

function [W, cache] = update_rmsprop(W, dLdW, lr, decayrate, cache)
% update_adam updates network weights W using RMSProp method
    cache = decayrate * cache + (1-decayrate) * dLdW.^2;
    W = W - lr * dLdW ./ (sqrt(cache) + 1e-10);
end

function [W, cache] = update_adagrad(W, dLdW, lr, cache)
% update_adam updates network weights W using AdaGrad method
    cache = cache + dLdW.^2;
    W = W - lr * dLdW ./ (sqrt(cache) + 1e-10);
end